; SOCZ80 CP/M 3 LDRBIOS (based on the MPM ldrbios 2013-11-19 Will Sowerbutts)
; Alan Cox 2014-07-01
; use zmac to assemble this
;
; zmac --rel ldrbios
;
; on CP/M
; link cpmldr[L100]=cpmldr,ldrbios
;
; Test by running the resulting binary on CP/M *unbanked*
;
; For production use link at the final desired address (remembering it will
; use 128 bytes below that as data), set the boot code to load it and off
; you go
;
; Currently supports RAMdisc booting. At some point it would be useful to
; roll a version that boots off the flash rom drive directly.
;

.z80 ; prefer z80 mnemonics

UART0_STATUS   equ 0x0000 ; [7: RX READY] [6: TX BUSY] [6 unused bits]
UART0_DATA     equ 0x0001
MMU_PAGESEL    equ 0xF8
MMU_DATA       equ 0xFA
MMU_PERM       equ 0xFB
MMU_FRAMEHI    equ 0xFC
MMU_FRAMELO    equ 0xFD
MMU_PTR_VAL0   equ 0xFC ; In page 17 mode
MMU_PTR_VAL1   equ 0xFD
MMU_PTR_VAL2   equ 0xFE
MMU_PTR_VAL3   equ 0xFF

; jump vector table used by CP/M
            jp boot         ; cold start
wboote:     jp bad          ; warm start
            jp bad	    ; console status
            jp bad          ; console character in
            jp conout       ; console character out
            jp bad          ; list character out
            jp bad          ; punch character out
            jp bad          ; reader character out
            jp home         ; move disk head to home position
            jp seldsk       ; select disk
            jp seltrk       ; set track number
            jp setsec       ; set sector number
            jp setdma       ; set DMA address
            jp read         ; read disk
            jp bad          ; write disk
            jp bad          ; return list status
            jp sectran      ; sector translate
            jp bad	    ; console output ready
            jp bad	    ; aux input is ready
            jp bad	    ; aux output is ready
            jp bad	    ; device table
            jp bad	    ; device init
            jp bad 	    ; drive table
            jp bad 	    ; multi block I/O
            jp bad	    ; flush data to disc
            jp bad	    ; block move (including banking)
            jp bad	    ; get current data and time into SCB
            jp bad	    ; select memory bank. Special: preserve regs
            jp bad	    ; set the bank for the next read/write sector
            jp bad	    ; set up for cross bank move
            jp bad	    ; extended BIOS
            jp 0	    ; reserved
            jp 0	    ; reserved


; CP/M 3 style disk parameter header
dpbase0:
            ; disk 0 (A)
            dw 0            ; sector translation table (0 = no translation)
	    db 0,0,0,0,0,0,0,0,0  ; BDOS scratch area
	    db 0            ; Media flag
            dw dpblk        ; DPB (disk parameter block)
            dw 0            ; CSV (unique scratch pad used to check for changed disks)
            dw alv00        ; ALV (unique scratch pad for allocation information)
	    dw dirbcb0	    ; Directory buffer control block
	    dw dtabcb       ; DTAB CB
            dw 0xFFFF       ; HASH
            db 0            ; HASH bank
            ; end of disk 0

dpbase1:
            ; disk 1 (B)
            dw 0            ; sector translation table (0 = no translation)
	    db 0,0,0,0,0,0,0,0,0  ; BDOS scratch area
	    db 0            ; Media flag
            dw dpblk        ; DPB (disk parameter block)
            dw 0            ; CSV (unique scratch pad used to check for changed disks)
            dw alv01        ; ALV (unique scratch pad for allocation information)
	    dw dirbcb1	    ; Directory buffer control block
	    dw dtabcb       ; DTAB CB
            dw 0xFFFF       ; HASH
            db 0            ; HASH bank
            ; end of disk 0


ndisks      equ 2           ; number of disks we defined

dirbcb0:    db 0xFF	    ; Disk Drive
            ds 3	    ; Record
            ds 1            ; Write Buffer Flag
            db 0            ; BDOS scratch byte
            ds 2            ; Track
            ds 2            ; Sector
            dw dirb0        ; Directory buffer
            db 0	    ; Bank
            dw 0	    ; Link

dirbcb1:    db 0xFF	    ; Disk Drive
            ds 3	    ; Record
            ds 1            ; Write Buffer Flag
            ds 1            ; BDOS scratch byte
            ds 2            ; Track
            ds 2            ; Sector
            dw dirb1        ; Directory buffer
            db 0	    ; Bank
            dw 0	    ; Link

dtabcb:     db 0xFF	    ; Disk Drive
            ds 3	    ; Record
            ds 1            ; Write Buffer Flag
            db 0            ; BDOS scratch byte
            ds 2            ; Track
            ds 2            ; Sector
            dw data0        ; Data buffer
            db 0	    ; Bank
            dw 0	    ; Link

; disk parameter block (can be shared by all disks with same configuration)
; my RAM disks are each 2MB in size;
;     track is 16384 bytes (128 x 128 sectors per track)
;     there are 128 tracks (128 x 128 x 128 = 2MB)
;     track 0 is reserved for the operating system, so there are 2048-16 = 2032 user sectors, and 16KB reserved for OS boot storage (which is plenty)
; using 1K sectors here resulted in a FUCK UP. don't do that! with 4K sectors it seems much happier.
dpblk:
            dw 128          ; SPT: number of 128 byte sectors per track
            db 5            ; BSH: block shift factor (see manual for table)
            db 31           ; BLM: block mask (see manual for table)
            db 1            ; EXM: extent mask (see manual for table, using entries marked N/A turns out to be a bad idea!)
            dw 508          ; DSM: (disk bytes / block bytes) - 1, change alv00/01 etc if you change this; this is the number of the last sector on the disk, excluding system tracks (ie more system tracks -> this gets smaller)
            dw 511          ; DRM: directory max entries - 1
            db 0xf0         ; AL0: directory sector allocation bitmask byte 0
            db 0x00         ; AL1: directory sector allocation bitmask byte 1
            dw 0x8000       ; CKS: check size (0x8000 = none, fixed)
            dw 1            ; OFF: track offset (number of system tracks)
	    db 0	    ; PSH - 128 byte sectors
	    db 0	    ; PHM - 128 byte sectors

; bios functions follow

boot:
	    ld hl, bootstr
            call strout
	    ret		    ; we have no special init needs
bad:
            ld hl, ouchithurts
            call strout
wefailed:   jr wefailed

conout:     ; write character from C to console
            in a, (UART0_STATUS)
            bit 6, a
            jr nz, conout ; loop again if transmitter is busy
            ld a, c
            out (UART0_DATA), a ; transmit character
            ret

seldsk:     ; select disk indicated by register C
       
            ld hl, 0    ; return code 0 indicates error
            ld a, c
            cp ndisks
            ret nc      ; return (with error code) if C >= ndisks ie illegal drive
            ld (curdisk), a ; store current disk
            ld hl, dpbase0
	    or a
	    ret z
	    ld hl, dpbase1
            ret

home:       
	    ld c, 0
            ; fall through into seltrk
seltrk:     ; set track given by register BC
            ld a, c
            ld (curtrack), a
            ret

setsec:     ; set sector given by register BC
            ld a, c
            ld (cursector), a
            ret

sectran:    ; logical to physical sector translation
            ; HL=BC ie 1:1 mapping (no translation)
            ld h, b
            ld l, c
            ret

devini:	    ; re-initialize device C, called on baud/mode change
            ; nothing to do for now
            ret

setdma:     ; set DMA address given by BC
            ld (curdmaaddr), bc ; may need to xfer to HL first?
            ret

read:       ; read from our RAM disk
            ld a, 0xff
            out (MMU_PAGESEL), a
            xor a
            out (MMU_PTR_VAL0), a
            ld a, (curdisk)     ; disk 1-3 maps to 2MB chunk 1 to 3
            inc a
            rrca                ; roll it round into bits 6:5 with 7 as 0
            rrca
            rrca
            ld b, a
            ld a, (curtrack)    ; bits 4-0 are track bits 6-2
            sra a               ; relies on <= 127 sectors/track
            sra a
            or b
            out (MMU_PTR_VAL1), a
            
            ld a, (curtrack)    ; we want the low bits of track but in the top
            rrca                ; move them into the top two bits
            rrca
            and 192             ; remove the chaff
            ld b, a
            ld a, (cursector)
            sra a               ; we know 0-127 so top bit will be 0, low
                                ; bit goes to carry
            or b                ;
            out (MMU_PTR_VAL2), a     ; high 6 bits of sector + bottom 2 of track

            ld a, (cursector)
            rrca		; recovery carry bit
            and 128
            out (MMU_PTR_VAL3), a     ; sector bottom bit

            ld hl, (curdmaaddr) ; Where to copy data
            ld bc, 128 * 256 + MMU_DATA; copy size, port
            inir
rw_done:    xor a ; return to bank 0
            ret

;---------------------------------------------------------------------------------------------------------------
; debug functions (ideally to be removed in final version, if we ever get that far!)
strout:     ; print string pointed to by HL
            ld a, (hl)
            cp 0
            ret z
            ld c, a
            call conout
            inc hl
            jr strout

ouchithurts: db "oops", 0
bootstr: db "Boot", 0
;---------------------------------------------------------------------------------------------------------------

; scratch RAM used by BIOS
curdisk:    db 0x11
curtrack:   db 0x22
cursector:  db 0x33
curdmaaddr: dw 0x4444
dirb0:      ds 128           ; directory scratch area
dirb1:      ds 128           ; directory scratch area
data0:	    ds 128	     ; data scratch area
alv00:      ds 64            ; allocation vector for disk 0, must be (DSM/8)+1 bytes
alv01:      ds 64            ; allocation vector for disk 1, must be (DSM/8)+1 bytes
